package graphlib;

/**
 *
 * <p>Title: </p>
 * <p>Description: Luokka matriisn ja vektoreiden laskutoimituksille. Vektoreita ksitelln
 * ykisulotteisina ja matriiseja kaksiulotteisina liukulukutaulukkoina.</p>
 * @author Annemari Auvinen
 */


public final class MatrixCalculations {
    public MatrixCalculations() {
    }
    
    /**
     * Alustaa vektorin alkioiksi nollan.
     * @param dim Vektorin koko
     * @return yksiulotteisen liukulukutaulukon
     */
    public static double[] initVector(int dim) {
        double[] vector = new double[dim];
        for(int i=0; i<dim; i++) {
            vector[i] = 0.0;
        }
        return vector;
    }
    
    
    /**
     * Vektoreiden yhteenlasku
     * @param v1  summattavan vektorin alkiot yksiulotteisessa taulukossa
     * @param v2  toisen summattavan vektorin alkiot yksiulottiesessa taulukossa
     * @return  Palauttaa yksiulotteisen liukulukutaulukon, johon laskettuna
     * parametreina tuotujen vektoreiden summa
     * @throws MatrixException
     */
    public static double[] sumVectors(double[] v1, double[] v2) throws MatrixException {
        if(v1.length!=v2.length)
            throw new MatrixException("Vektoreissa eri mr alkioita, ei voida summata");
        int dim = v1.length;
        double[] vector = initVector(dim);
        for(int i=0; i<dim; i++) {
            vector[i] = v1[i]+v2[i];
        }
        return vector;
    }
    
    /**
     * Vektoreiden vhennyslasku
     * @param v1  vhennettvn vektorin alkiot yksiulotteisessa taulukossa
     * @param v2  vhentjvektorin alkiot yksiulottiesessa taulukossa :-)
     * @return  Palauttaa yksiulotteisen liukulukutaulukon, johon laskettuna
     * parametreina tuotujen vektoreiden erotus
     * @throws MatrixException
     */
    public static double[] subtractionOfVectors(double[] v1, double[] v2) throws MatrixException {
        if(v1.length!=v2.length)
            throw new MatrixException("Vektoreissa eri mr alkioita, ei voida vhent");
        int dim = v1.length;
        double[] vector = initVector(dim);
        for(int i=0; i<dim; i++) {
            vector[i] = v1[i]-v2[i];
        }
        return vector;
    }
    
    /**
     * Vektorin kertominen reaaliluvulla
     * @param v1 Kerrottavan vektorin alkiot yksiulotteisessa taulukossa
     * @param x  Reaalilukukerroin
     * @return  Palauttaa ykisulotteisen liukulukutaulukon, johon laskettuna
     * tuodun vektorin ja reaaliluvun tulo
     */
    public static double[] multiplyVectorByX(double[] v1, double x) {
        int dim = v1.length;
        double[] vector = initVector(dim);
        for(int i=0; i<dim; i++) {
            vector[i] = v1[i]*x;
        }
        return vector;
    }
    
    /**
     * Vektorien sistulo
     * @param v1 Kerrottavan vektorin alkiot yksiulotteisessa taulukossa
     * @param v2 Toisen kerrottavan vektorin alkiot yksiulotteisessa taulukossa
     * @return  Vektoreiden sistulon
     * @throws MatrixException
     */
    public static double innerProduct(double[] v1, double[] v2) throws MatrixException{
        int dim = v1.length;
        double product = 0.0;
        if(v1.length !=  v2.length)
            throw new MatrixException("Vektoreissa eri mr alkioita, ei voida laskea sistuloa");
        for(int i=0; i<dim; i++) {
            product += v1[i]*v2[i];
        }
        return product;
    }
    
    public static double length(double[] v) throws MatrixException {
        return Math.sqrt(innerProduct(v, v));
    }
    
    /**
     * Kolmiulotteisten vektoreiden ristitulo
     * @param v1 Kerrottavan vektorin alkiot yksiulotteisessa taulukossa
     * @param v2 Toisen kerrottavan vektorin alkiot yksiulotteisessa taulukossa
     * @return Yksiulotteisen taulukon, johon laskettuna vektoreiden ristitulo
     * @throws MatrixException
     */
    public static double[] crossProduction(double[] v1, double[] v2) throws MatrixException {
        if(v1.length != 3 || v2.length != 3)
            throw new MatrixException("Vektorit eivt kolmiulotteisia, ei voida laskea ristituloa");
        double[] vector = initVector(3);
        vector[0] = v1[1]*v2[2]-v1[2]*v2[1];
        vector[1] = v1[2]*v2[0]-v1[0]*v2[2];
        vector[2] = v1[0]*v2[1]-v1[1]*v2[0];
        return vector;
    }
    
    /**
     * Alustaa matriisin
     * @param dim1 Matriisin rivien lukumr
     * @param dim2 Matriisin sarakkeiden lukumr
     * @return Palauttaa dim1*dim2 -kokoisen matriisin, jonka alkioina nolla
     */
    public static double[][] initMatrix(int dim1, int dim2) {
        double[][] matrix = new double[dim1][dim2];
        for(int i=0; i<dim1; i++) {
            for(int j=0; j<dim2; j++) {
                matrix[i][j] = 0.0;
            }
        }
        return matrix;
    }
    
    /**
     * Matriisien yhteenlasku
     * @param m1 Yhteenlaskettava matriisi
     * @param m2 Toinen yhteenlaskettava matriisi
     * @return Palauttaa matriisin, johon summa laskettuna
     * @throws MatrixException
     */
    public static double[][] sumMatrixs(double[][] m1, double[][] m2) throws MatrixException {
        int rows = m1.length;
        int columns = m1[0].length;
        if(m2.length != rows || m2[0].length != columns)
            throw new MatrixException("Matriiseissa eri mr alkioita, ei voida summata");
        double[][] sum = initMatrix(rows, columns);
        System.out.println(m1.length);
        for(int i=0; i<rows; i++) {
            for(int j=0; j<columns; j++) {
                sum[i][j] = m1[i][j]+m2[i][j];
            }
        }
        return sum;
        
    }
    
    /**
     * Matriisin kertoinen reaaliluvulla
     * @param m1 Kerrottava matriisi
     * @param x Reaalilukukerroin
     * @return Matriisin, johon tulo on laskettu
     */
    public static double[][] multiplyMatrixByX(double[][] m1, double x) {
        int rows = m1.length;
        int columns = m1[0].length;
        double[][] matrix = initMatrix(rows, columns);
        for(int i=0; i<rows; i++) {
            for(int j=0; j<columns; j++) {
                matrix[i][j] = m1[i][j]*x;
            }
        }
        return matrix;
    }
    
    
    
    /**
     * Matriisin kertominen vektorilla
     * @param m Kerrottava matriisi
     * @param v Vektori, jolla kerrotaan
     * @return Vektorin, johon tulo laskettu
     * @throws MatrixException Mikli matriisissa ei ole yht monta saraketta kuin
     * vektorissa on alkioita.
     */
    public static double[] multiplyMatrixByVector(double[][] m, double[] v) throws MatrixException{
        int rows = m.length;
        int columns = m[0].length;
        int vDim= v.length;
        if(columns != vDim)
            throw new MatrixException("Matriisin sarakkeiden ja vektorin alkioiden mr ei tsm, ei voida kertoa keskenn");
        double[] vector = initVector(vDim);
        for(int i=0; i<rows; i++) {
            double sum = 0.0;
            for(int j=0; j<columns; j++) {
                sum += m[i][j]*v[j];
            }
            vector[i] = sum;
        }
        return vector;
    }
    
    /**
     * Matriisitulo
     * @param m1 Kerrottava matriisi
     * @param m2 Toinen kerrottava matriisi
     * @return Matriisn, johon tulo on laskettu
     * @throws MatrixException Mikli matriisissa m1 ei ole yht monta saraketta kuin
     * matriisissa m2 on rivej
     */
    public static double[][] multiplyMatrixByMatrix(double[][] m1, double[][] m2) throws MatrixException{
        int dim1 = m1.length;
        int dim2 = m1[0].length;
        if(dim2 != m2.length)
            throw new MatrixException("Matriisien kertolasku ei onnistu, sarakkeiden ja rivien lukumr ei tsm");
        int dim3 = m2[0].length;
        double[][] product = initMatrix(dim1, dim3 );
        for(int i=0; i<dim1; i++) {
            for(int j=0; j<dim3; j++) {
                double sum=0.0;
                for(int k=0; k<dim2; k++) {
                    sum +=  m1[i][k]*m2[k][j];
                }
                product[i][j] = sum;
            }
        }
        return product;
        
    }
    
    /**
     * Matriisin transpoosi
     * @param m Transponoitava matriisi
     * @return Matriisin, joka on parametrina tuodun matriisin transpoosi
     */
    public static double[][] transpose(double[][] m){
        int dim1 = m.length;
        int dim2 = m[0].length;
        double[][] transpose = initMatrix(dim2, dim1);
        for(int i=0; i<dim1; i++) {
            for(int j=0; j<dim2; j++) {
                transpose[j][i] = m[i][j];
            }
        }
        return transpose;
    }
    
    /**
     * Tulostaa matriisin alkiot nytlle
     * @param m Tulostettava matriisi
     */
    public static void printMatrix(double[][] m) {
        System.out.println("Matrix: ");
        for(int i=0; i<m.length; i++) {
            for(int j=0; j<m[0].length; j++) {
                System.out.print(m[i][j]+" ");
            }
            System.out.println();
        }
    }
    
    /**
     * Tulostaa vektorin alkiot nytlle
     * @param v Tulostettava vektori
     */
    public static void printVector(double[] v) {
        System.out.println("Vector: ");
        int dim = v.length;
        for(int i=0; i<dim; i++) {
            System.out.print(v[i]+" ");
        }
        System.out.println();
        
    }
    
    public static double [][] R(double a) {
        return new double [][] {
            {Math.cos(a), -Math.sin(a), 0},
            {Math.sin(a), Math.cos(a),  0},
            {0,           0,            1}};
    }
    
    public static double[] normalize(double[] q) throws MatrixException {
        double len = length(q);
        for(int i = 0; i < q.length; i++) {
            q[i] = q[i]/len;
        }
        return q;
    }
    
    public static double [][] unitMatrix() {
        return new double [][] {
            {1, 0, 0},
            {0, 1,  0},
            {0, 0,  1}};
    }
    
    public static double[][] star(double[] a) throws MatrixException {
        if(a.length!= 3)
            throw new MatrixException("Length of the array should be 3.");
        return new double[][] {
            {0, -a[2], a[1]},
            {a[2], 0, -a[0]},
            {-a[1], a[0], 0}};
    }
    
    public static double[][] quaternionToMatrix(double[] q) throws MatrixException {
        if(q.length != 4)
            throw new MatrixException("Length of the array should be 4.");
        return new double[][] {
            {1-2*q[2]*q[2]-2*q[3]*q[3], 2*q[1]*q[2]-2*q[0]*q[3], 2*q[1]*q[3]+2*q[0]*q[2]},
            {2*q[1]*q[2]+2*q[0]*q[3], 1-2*q[1]*q[1]-2*q[3]*q[3], 2*q[2]*q[3]-2*q[0]*q[1]},
            {2*q[1]*q[3]-2*q[0]*q[2], 2*q[2]*q[3]+2*q[0]*q[1], 1-2*q[1]*q[1]-2*q[2]*q[2]}};
    }
    
    public static double[] quaternionToMatrix4DVector(double[] q) throws MatrixException {
        if(q.length != 4)
            throw new MatrixException("Length of the array should be 4.");
        return new double[] {
            1-2*q[2]*q[2]-2*q[3]*q[3], 2*q[1]*q[2]-2*q[0]*q[3], 2*q[1]*q[3]+2*q[0]*q[2], 0,
            2*q[1]*q[2]+2*q[0]*q[3], 1-2*q[1]*q[1]-2*q[3]*q[3], 2*q[2]*q[3]-2*q[0]*q[1], 0,
            2*q[1]*q[3]-2*q[0]*q[2], 2*q[2]*q[3]+2*q[0]*q[1], 1-2*q[1]*q[1]-2*q[2]*q[2], 0,
            0, 0, 0, 1};
    }
    
    public static double[] quaternionProduct(double[] q1, double[] q2) throws MatrixException {
        if(q1.length != 4 || q1.length != 4)
            throw new MatrixException("Length of the array should be 4.");
        double [] q3 = new double [4];
        
        double s1 = q1[0];
        double s2 = q2[0];
        double [] v1 = new double [] { q1[1], q1[2], q1[3] };
        double [] v2 = new double [] { q2[1], q2[2], q2[3] };
        
        q3[0] = s1 * s2 - innerProduct(v1, v2);
        double [] v3 = sumVectors(sumVectors(multiplyVectorByX(v2, s1), multiplyVectorByX(v1, s2)), 
        crossProduction(v1, v2));
        
        q3[1] = v3[0];
        q3[2] = v3[1];
        q3[3] = v3[2];
        
        return q3;
    }
    
    public static double calculateDistance( double[] p1, double[] p2) throws MatrixException {
      if(p1.length != p2.length || p1.length != 3)
         throw new MatrixException("Lengths of the arrays should be 3.");
      double dist = Math.sqrt(Math.pow((p2[2]-p1[2]), 2.0) + Math.pow((p2[1]-p1[1]), 2.0) + Math.pow((p2[0]-p1[0]), 2.0));
      return dist;
    
    }
    
}
