/*
 * Collision.java
 *
 * Created on 28. lokakuuta 2004, 9:44
 */

package marbles;

import graphlib.*;

/**
 *
 * @author  annauvi
 */
public class Contact {
    private Marble m1 = null;
    private Marble m2 = null;
    double[] p = new double[3]; //trmyspiste
    double[] n = new double[3]; //marble2:n normaali
    private final double PLANE_FRICTION = 0.2;
    private final double MARBLE_FRICTION = 0.05;

    
    
    /** Creates a new instance of Collision */
    public Contact(Marble m1, Marble m2, double[] p, double[] n) {
        this.m1 = m1;
        this.m2 = m2;
        this.p = p;
        this.n = n;
    }
    
    public Contact( Marble m, double[] p, double[] n) {
        this.m1 = m;
        this.p = p;
        this.n = n;
    }
    
    public Marble getMarble1() {
        return m1;
    }
    
    public Marble getMarble2() {
        return m2;
    }
    
    public double[] getP() {
        return p;
    }
    
    public double[] getN() {
        return n;
    }
    
    //Laskee normaalivektorin derivaatan
    public double[] computeNdot() throws MatrixException {
        //jos tason kanssa, niin eik silloin vertex/face kontakti??
        //jos kappale B on fiksattu, niin derivaatta on nolla, eik??
        if(m2 == null) {
            return MatrixCalculations.initVector(3);
        }
        else {
            double[] ea = MatrixCalculations.multiplyVectorByX(getN(), -1.0);
            double[] eb = getN();
            double[] eadot = MatrixCalculations.crossProduction(m1.getOmega(), ea);
            double[] ebdot = MatrixCalculations.crossProduction(m2.getOmega(), eb);
            double[] n = MatrixCalculations.crossProduction(ea, eb);
            double[] z = MatrixCalculations.sumVectors(MatrixCalculations.crossProduction(eadot, eb), MatrixCalculations.crossProduction(ea, ebdot));
            double l = MatrixCalculations.length(n);
            if(l== 0)
                return n;
            n = MatrixCalculations.multiplyVectorByX(n, 1.0/l);
            return MatrixCalculations.multiplyVectorByX(MatrixCalculations.subtractionOfVectors(z, MatrixCalculations.crossProduction(MatrixCalculations.crossProduction(z, n), n)), 1.0/l);
        }
        
    }
    
    /**
     * Marble m1 - colliding Marble
     * Marble m2 - colliding Marble
     * double [] p - collision point
     * double e - epsilon
     */
    private void collisionBetweenMarbles(double e) throws MatrixException {
        double[] padot = m1.getPointVelocity(p);
        double[] pbdot = m2.getPointVelocity(p);
        double[] ra = MatrixCalculations.subtractionOfVectors(p, m1.getX());
        double[] rb = MatrixCalculations.subtractionOfVectors(p, m2.getX());
        double vrel = MatrixCalculations.innerProduct(n, MatrixCalculations.subtractionOfVectors(padot, pbdot));
        double numerator = -1 * (1 + e) * vrel;
        
        //denominators
        double term1 = 1.0/m1.getMass();
        double term2 = 1.0/m2.getMass();
        double[] term31 = MatrixCalculations.crossProduction(ra, n);
        double[] term32 = MatrixCalculations.multiplyMatrixByVector(m1.getIinv(), term31);
        double term3 = MatrixCalculations.innerProduct(n, MatrixCalculations.crossProduction(term32,  ra));
        double[] term41 = MatrixCalculations.crossProduction(rb, n);
        double[] term42 = MatrixCalculations.multiplyMatrixByVector(m2.getIinv(), term41);
        double term4 = MatrixCalculations.innerProduct(n, MatrixCalculations.crossProduction(term42,  rb));
        //impulse magnitude
        double j = numerator/(term1 + term2 + term3 + term4);
        //Impulse
        double [] J = MatrixCalculations.multiplyVectorByX(n, j);
        // Virheen generointi...
        double rand = Math.random() / 10;
        double tmp = Math.random();
        if (tmp < 0.5) rand = -rand;
        tmp = Math.random();
        if (tmp < 0.5) J[0] += rand * e;
        else J[1] += rand * e;
        
        m1.addImpulse(J, ra);
        m2.addImpulse(MatrixCalculations.multiplyVectorByX(J, -1.0), rb);

        /* Ei toimi, hukkaa pallot kun kitka menee rettmn :P
        // Fake friction
        // padot projektio tangenttitasolle padot - (padot,ra)ra
        double [] afriction = MatrixCalculations.subtractionOfVectors(padot,
        MatrixCalculations.multiplyVectorByX(ra, MatrixCalculations.innerProduct(padot, ra)));
        afriction = MatrixCalculations.multiplyVectorByX(afriction, 
        -0.1 * MatrixCalculations.length(padot)); // Negaatio * hattuvakio
        System.out.println("padot " + padot[0] + ", " + padot[1] + ", " + padot[2]);
        System.out.println("A friction " + afriction[0] + ", " + afriction[1] + ", " + afriction[2]);

        double [] bfriction = MatrixCalculations.subtractionOfVectors(pbdot,
        MatrixCalculations.multiplyVectorByX(rb, MatrixCalculations.innerProduct(pbdot, rb)));
        bfriction = MatrixCalculations.multiplyVectorByX(bfriction, 
        -0.1 * MatrixCalculations.length(pbdot)); // Negaatio * hattuvakio
        System.out.println("pbdot " + pbdot[0] + ", " + pbdot[1] + ", " + pbdot[2]);
        System.out.println("B friction " + bfriction[0] + ", " + bfriction[1] + ", " + bfriction[2]);

        m1.addL(afriction, ra); // Lis vain kulmaliikemr & kulmakiihtyvyytt
        m2.addL(bfriction, rb); // Lis vain kulmaliikemr & kulmakiihtyvyytt
         */
}
    
    /**
     * Marble m - colliding Marble
     * double [] p - collision point
     * double e - epsilon
     */
    private void collisionWithPlane(double e) throws MatrixException {
        double [] padot = m1.getPointVelocity(p);
        // double [] pbdot = MatrixCalculations.initVector(3);
        // Vektori keskipisteest trmyspisteeseen
        double [] ra = MatrixCalculations.subtractionOfVectors(p, m1.getX());
        // Relative velocity -> IF <0 collide; ==0 contact; >0 moving apart
        double vrel = MatrixCalculations.innerProduct(n, padot);
        double numerator = -1 * (1 + e) * vrel;
        //double numerator = -1 * (1+e) * vrel;
        // Denominators
        double term1 = 1.0 / m1.getMass();
        double [] term21 = MatrixCalculations.crossProduction(ra, n);
        double [] term22 = MatrixCalculations.multiplyMatrixByVector(m1.getIinv(), term21);
        double term2 = MatrixCalculations.innerProduct(n, MatrixCalculations.crossProduction(term22,  ra));
        // Impulse magnitude
        double j = numerator/(term1 + term2);
        // Impulse
        double [] J = MatrixCalculations.multiplyVectorByX(n, j);
        m1.addImpulse(J, ra);
        // "Fake friction"
        double [] friction = MatrixCalculations.multiplyVectorByX(m1.getV(), -1*PLANE_FRICTION);
        friction[2] = 0; // Projektio tasolle
        // Skaalaus
        // rettmn meno hvitt kuulia
        if (MatrixCalculations.length(friction) > 10000)
            friction = MatrixCalculations.multiplyVectorByX(friction, 10000/ MatrixCalculations.length(friction));
        m1.setL(friction, ra); // Muuttaa vain kulmaliikemr & kulmakiihtyvyytt
        // Ei vaikuta muuhun liikemrn

    }
    
    public void collision(double e) throws MatrixException {
        if (m2 == null)
            collisionWithPlane(e);
        else
            collisionBetweenMarbles(e);
    }
    
}
