/*
 * MarbleContainer.java
 *
 * Created on 19. heinkuuta 2004, 15:50
 */

package marbles;

import graphlib.*;
import marbles.force.*;

import java.util.*;

/**
 *
 * @author  millat
 */
public class MarbleContainer implements Derivable {
    
    private final double COLLISION_EPSILON = 1.0 / 300.0;
    private final double BOUNCE_EPSILON = 0.9;
    private final double CONTACT_EPSILON = 1.0 / 3.0;
    
    private Vector marbles = new Vector();
    private Vector forces = new Vector();
    private Display display = null;
    private DESolver solver = null;
    private Misc misc = new Misc();
    
    private boolean animationRunning = false;
    
    /** Creates a new instance of MarbleContainer */
    public MarbleContainer(int cubeSize, double cubeDistance, double marbleRadius, double marbleMass) {
        double z = cubeDistance + marbleRadius; // Alin rivi
        for (int i = 0; i < cubeSize; i++) {
            double x = - cubeSize * marbleRadius;
            for (int j = 0; j < cubeSize; j++) {
                double y = - cubeSize * marbleRadius;
                for (int k = 0; k < cubeSize; k++) {
                    marbles.add(new Marble(new double [] { x, y, z }, marbleRadius, marbleMass));
                    y += 2 * marbleRadius;
                }
                x += 2 * marbleRadius;
            }
            z += 2 * marbleRadius;
        }
      /*
    switch (cubeSize) {
      case 1: {
        marbles.add(new Marble(new double [] { 0, 0, cubeDistance + marbleRadius }, marbleRadius, marbleMass));
        break;
      }
      case 2: {
        marbles.add(new Marble(new double [] { marbleRadius, marbleRadius, cubeDistance + marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { -marbleRadius, marbleRadius, cubeDistance + marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { marbleRadius, -marbleRadius, cubeDistance + marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { -marbleRadius, -marbleRadius, cubeDistance + marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { marbleRadius + 0.1, marbleRadius + 0.1, cubeDistance + marbleRadius + 2 * marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { -marbleRadius, marbleRadius, cubeDistance + marbleRadius + 2 * marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { marbleRadius, -marbleRadius, cubeDistance + marbleRadius + 2 * marbleRadius }, marbleRadius, marbleMass));
        marbles.add(new Marble(new double [] { -marbleRadius, -marbleRadius, cubeDistance + marbleRadius + 6 * marbleRadius }, marbleRadius, marbleMass));
        break;
      }
    }
       */
        this.solver = new MidpointSolver(this);
        this.forces.add(new ForceGravity());
    }
    
    public void setDisplay(Display d) {
        this.display = d;
    }
    
    public int getMarbleCount() {
        return marbles.size();
    }
    
    public boolean isAnimationRunning() {
        return animationRunning;
    }
    
    public Marble getMarble(int index) throws Exception {
        if (index < 0 || index >= marbles.size())
            throw new Exception("Bad index");
        return (Marble) marbles.get(index);
    }
    
    public double [] marblesToArray() {
        // Kopioi kaikkien kuulien taulukot yhteen
        double [] y = new double [getMarbleCount() * Marble.STATE_SIZE];
        int yIndex = 0;
        for(int i=0; i<getMarbleCount(); i++) {
            double [] tmp = ((Marble)marbles.get(i)).stateToArray();
            for (int j = 0; j < tmp.length; j++)
                y[yIndex++] = tmp[j];
        }
        
        return y;
    }
    
    public void arrayToMarbles(double [] y) throws MatrixException {
        // Kopioi taulukosta arvot kuuliin
        //meneekhn tmp:hen sijoitus oikein..
        int yIndex=0;
        double[] tmp = new double[Marble.STATE_SIZE];
        for(int i=0; i<getMarbleCount(); i++) {
            for(int j=0; j<tmp.length; j++) {
                tmp[j]=y[yIndex++];
            }
            ((Marble)marbles.get(i)).arrayToState(tmp);
        }
    }
    
    /** Compute force and torque acting on each marble */
    public void computeForceAndTorque(double t, int marbleIndex) throws MatrixException {
        // Voimien laskenta, painovoima, kitka, ilmanvastus tms
        // Tulokset jokaisen Marblen muuttujiin force ja torque
        Marble m = (Marble)marbles.get(marbleIndex);
        m.clearForce();
        for (Enumeration e = forces.elements(); e.hasMoreElements();) {
            Force f = (Force) e.nextElement();
            f.applyForce(m);
        }
    }
    
    
    /**
     * Vector contactVector Vector including the Contacts
     */
    
    public void contact(Vector contacts) throws MatrixException {
       // System.out.println("Kontakti. "+contacts.size());
        //Contact[] contacts = (Contact[])contactVector.toArray();
        //lasketaan kertoimet aij ja bij
        double[][] aMat = computeAMatrix(contacts);
        double[] bVec = computeBVector(contacts);
        //Ratkaistaan fj:t
  //      System.out.println("Fvec:");
        double[] fVec = misc.solveF(aMat, bVec);
   /*     for(int i=0; i<contacts.size(); i++) {
            System.out.println(fVec[i]);
        }*/
        //Listn kontaktivoimat kuulien force- ja torque-kenttiin, toiseen positiivisena ja toiseen negatiivisena
        for(int i=0; i<contacts.size(); i++) {
            double f = fVec[i];
            Contact contact = (Contact)contacts.get(i);
            double[] n = contact.getN();
            Marble a = contact.getMarble1();
            Marble b = contact.getMarble2();
            double [] forceToAdd =  MatrixCalculations.multiplyVectorByX(n,f);
            double[] pa = MatrixCalculations.subtractionOfVectors(contact.getP(), a.getX());
          //  System.out.println("Listn a:han: "+forceToAdd[0]+" "+forceToAdd[1]+" "+forceToAdd[2]);
            a.addContactForce(forceToAdd, pa); //lis voiman ja torquen
            //esimerkissa torquelaskussa oli sistulo!!
           // a.addTorque(MatrixCalculations.crossProduction(MatrixCalculations.subtractionOfVectors(contact.getP(), a.getX()), forceToAdd));
            
            //mikli toinen osapuoli ei ole taso,
            if(b!= null) {
                double[] pb = MatrixCalculations.subtractionOfVectors(contact.getP(), b.getX());
                b.addContactForce(MatrixCalculations.multiplyVectorByX(forceToAdd, -1.0), pb);
             //   b.addTorque(MatrixCalculations.multiplyVectorByX(MatrixCalculations.crossProduction(MatrixCalculations.subtractionOfVectors(contact.getP(), b.getX()), forceToAdd), -1.0));
            }
        }
    }
    
    // Tarvitaan kontaktin laskemisessa
    public double[][] computeAMatrix(Vector contacts) throws MatrixException {
        double[][] aMat = new double[contacts.size()][contacts.size()];
        for(int i=0; i<contacts.size(); i++) {
            for(int j=0; j<contacts.size(); j++) {
                //   System.out.println(i+" "+j);
                aMat[i][j] = computeAij((Contact)contacts.get(i), (Contact)contacts.get(j));
            }
        }
        return aMat;
    }
    
    
    public double computeAij(Contact ci, Contact cj) throws MatrixException {
        //jos kontakteissa olevat kappaleet ovat erit, palautetaan nolla??
        if((ci.getMarble1() != cj.getMarble1()) && (ci.getMarble2() != cj.getMarble2()) && (ci.getMarble1() != cj.getMarble2()) && (ci.getMarble2() != cj.getMarble1()))
            return 0.0;
        Marble a = ci.getMarble1();
        Marble b = ci.getMarble2();
        double[] ni = ci.getN();
        double[] nj = cj.getN();
        double[] pi = ci.getP();
        double[] pj = cj.getP();
        double[] ra = MatrixCalculations.subtractionOfVectors(pi, a.getX());
        double[] rb = MatrixCalculations.initVector(3);
        //jos toinen kappale on mys kuula
        if(b != null) {
            rb = MatrixCalculations.subtractionOfVectors(pi, b.getX());
        }
        
        double[] forceOnA = MatrixCalculations.initVector(3);
        double[] torqueOnA = MatrixCalculations.initVector(3);
        double[] forceOnB = MatrixCalculations.initVector(3);
        double[] torqueOnB = MatrixCalculations.initVector(3);
        
        //Kuulaan a vaikuttavat voimat
        if(cj.getMarble1() == ci.getMarble1()) {
            forceOnA = nj;
            torqueOnA = MatrixCalculations.crossProduction(MatrixCalculations.subtractionOfVectors(pj, a.getX()), nj);
        }
        else if (cj.getMarble2() == ci.getMarble1()) {
            forceOnA = MatrixCalculations.multiplyVectorByX(nj, -1.0);
            torqueOnA = MatrixCalculations.crossProduction(MatrixCalculations.subtractionOfVectors(pj, a.getX()), nj);
        }
        
        //Kuulaan b vaikuttavat voimat
        if(cj.getMarble1() == ci.getMarble2()) {
            forceOnB = nj;
            torqueOnB = MatrixCalculations.crossProduction(MatrixCalculations.subtractionOfVectors(pj, b.getX()), nj);
        }
        else if((cj.getMarble2() == ci.getMarble2()) && cj.getMarble2() != null ) {
            forceOnB = MatrixCalculations.multiplyVectorByX(nj, -1.0);
            torqueOnB = MatrixCalculations.crossProduction(MatrixCalculations.subtractionOfVectors(pj, b.getX()), nj);
        }
        
        //Kontaktin vaikutus kuulien  lineaariseen ja kulmakiihtyvyyteen (nopeuteen?)
        double[] aLinear = MatrixCalculations.multiplyVectorByX(forceOnA, 1.0/a.getMass());
        double[] aAngular = MatrixCalculations.crossProduction(MatrixCalculations.multiplyMatrixByVector(a.getIinv(), torqueOnA), ra);
        double[] bAngular = MatrixCalculations.initVector(3);
        double[] bLinear = MatrixCalculations.initVector(3);
        
        //jos toinen kappale on kuula
        if(b!= null) {
            bAngular = MatrixCalculations.crossProduction(MatrixCalculations.multiplyMatrixByVector(b.getIinv(), torqueOnB), rb);
            bLinear = MatrixCalculations.multiplyVectorByX(forceOnB, 1.0/b.getMass());
        }
        double[] distinct = MatrixCalculations.subtractionOfVectors(MatrixCalculations.sumVectors(aLinear, aAngular), MatrixCalculations.sumVectors(bLinear, bAngular));
        return MatrixCalculations.innerProduct(ni, distinct);
    }
    
    //Tarvitaan kontaktien laskemisessa
    public double[] computeBVector(Vector contacts) throws MatrixException {
        double[] bVec = new double[contacts.size()];
        for(int i=0; i<contacts.size(); i++) {
            Contact c = (Contact)contacts.get(i);
            Marble a = c.getMarble1();
            Marble b = c.getMarble2();
            double[] n = c.getN();
            double[] ra = MatrixCalculations.subtractionOfVectors(c.getP(), a.getX());
            double[] rb = MatrixCalculations.initVector(3);
            
            //Kuulien nykyiset voimat
         //   double[] fExtA = a.getForce();
         //   double[] tExtA = a.getTorque();
            
          //  double[] fExtB = MatrixCalculations.initVector(3);
          //  double[] tExtB = MatrixCalculations.initVector(3);
            
            //Jos toinen kappale on mys kuula
            if(b!=null) {
                rb = MatrixCalculations.subtractionOfVectors(c.getP(), b.getX());
            //    fExtB = b.getForce();
            //    tExtB = b.getTorque();
            }
            
            //lasketaan osissa p(t), ensin ulkoisten voimien osalta ja sitten nopeuden osalta kummallekin kappaleelle,
            //tasolle arvot ovat nollia
            double[] aSum = MatrixCalculations.sumVectors(a.getAcceleration(),
            MatrixCalculations.crossProduction(a.getAngularAcceleration(), ra));
           
          //  double[] aExtPart = MatrixCalculations.sumVectors(MatrixCalculations.multiplyVectorByX(fExtA, 1.0/a.getMass()),
          //  MatrixCalculations.crossProduction(MatrixCalculations.multiplyMatrixByVector(a.getIinv(), tExtA), ra));
            double[] bSum = MatrixCalculations.initVector(3);
          //  double[] bExtPart = MatrixCalculations.initVector(3);

            if(b!= null) {
                bSum = MatrixCalculations.sumVectors(b.getAcceleration(), 
                MatrixCalculations.crossProduction(b.getAngularAcceleration(), rb));
            //  bExtPart = MatrixCalculations.sumVectors(MatrixCalculations.multiplyVectorByX(fExtB, 1.0/b.getMass()),
            //    MatrixCalculations.crossProduction(MatrixCalculations.multiplyMatrixByVector(b.getIinv(), tExtB), rb));

            }
            
            aSum = MatrixCalculations.sumVectors(aSum, MatrixCalculations.crossProduction(a.getOmega(),MatrixCalculations.crossProduction(a.getOmega(), ra)));
           /*double[] aVelPart1 = MatrixCalculations.crossProduction(a.getOmega(),MatrixCalculations.crossProduction(a.getOmega(), ra));
           double[] aVelPart2 = MatrixCalculations.crossProduction(MatrixCalculations.multiplyMatrixByVector(a.getIinv(), MatrixCalculations.crossProduction(a.getL(), a.getOmega())), ra);
           double[] aVelPart = MatrixCalculations.sumVectors(aVelPart1, aVelPart2);
*/
            
           // double[] bVelPart =  MatrixCalculations.initVector(3);
            if(b!=null) {
                bSum = MatrixCalculations.sumVectors(bSum, MatrixCalculations.crossProduction(b.getOmega(),MatrixCalculations.crossProduction(b.getOmega(), rb)));
               /* double[] bVelPart1 = MatrixCalculations.crossProduction(b.getOmega(),MatrixCalculations.crossProduction(b.getOmega(), rb));
                double[] bVelPart2 = MatrixCalculations.crossProduction(MatrixCalculations.multiplyMatrixByVector(b.getIinv(), MatrixCalculations.crossProduction(b.getL(), b.getOmega())), rb);
                bVelPart = MatrixCalculations.sumVectors(bVelPart1, bVelPart2);
*/
            }
  //          double[] aSum = MatrixCalculations.sumVectors(aExtPart, aVelPart);
  //          double[] bSum = MatrixCalculations.sumVectors(bExtPart, bVelPart);

            double k1 = MatrixCalculations.innerProduct(n, MatrixCalculations.subtractionOfVectors(aSum, bSum));
            
            //normaalivektorin derivaatta
            double[] ndot = c.computeNdot();
            
            
            double[] bVelocity = MatrixCalculations.initVector(3);
            if(b!=null)
                bVelocity = b.getPointVelocity(c.getP());
            double k2 = MatrixCalculations.innerProduct(MatrixCalculations.multiplyVectorByX(ndot, 2),MatrixCalculations.subtractionOfVectors(a.getPointVelocity(c.getP()),bVelocity));
            bVec[i] = k1 + k2;

        }
        return bVec;
    }
    
    
    
    
    
    public double [] dydt(double [] y, double t) throws MatrixException {
        double [] yDot = new double [y.length];
        int yDotIndex = 0;
        
        arrayToMarbles(y);
        
        for (int i = 0; i < getMarbleCount(); i++) {
            computeForceAndTorque(t, i);
            double [] tmp = ((Marble) marbles.get(i)).ddtStateToArray();
            for (int j = 0; j < tmp.length; j++)
                yDot[yDotIndex++] = tmp[j];
        }
        return yDot;
    }
    
    /**
     * Returns Marbles which collide now
     */
    private Vector [] checkCollisions() throws MatrixException {
        Vector collisions = new Vector();
        Vector contacts = new Vector();
        Vector [] all = new Vector [] { collisions, contacts };
        // Tason kanssa
        for (int i = 0; i < marbles.size(); i++) {
            Marble m = (Marble)marbles.get(i);
            if (m.getX(2) <= (m.getRadius() + COLLISION_EPSILON)) { //tarviiko yhtsuuruutta, tuleehan kontaktit mukaan..??
                double [] pDot = m.getPointVelocity(new double [] { m.getX(0), m.getX(1), 0.0 });
                // Kontakti
              //  if(pDot[2] < 0 && pDot[2] > -1*CONTACT_EPSILON) {
                if(pDot[2] == 0) {
                //    contacts.add(new Contact(m, new double [] { m.getX(0), m.getX(1), 0.0 }, new double[] {0.0, 0.0, 1.0} ));
                    // Trmys
                //} else if (pDot[2] < -1*COLLISION_EPSILON) {
                } else if (pDot[2] < 0) {
                    collisions.add(new Contact(m, new double [] { m.getX(0), m.getX(1), 0.0 }, new double[] {0.0, 0.0, 1.0} ));
                }
            }
        }
        
        for(int i=0; i<marbles.size(); i++) {
            Marble m1 = (Marble)marbles.get(i);
            for(int j=i+1; j<marbles.size(); j++) {
                Marble m2 = (Marble)marbles.get(j);
                double distance = MatrixCalculations.calculateDistance(m2.getX(), m1.getX());
                if(distance <= (m1.getRadius()+m2.getRadius()+COLLISION_EPSILON)) {
                    //normaali m2:sesta m1:en suuntaan
                    double[] vFromM2toM1 = MatrixCalculations.subtractionOfVectors(m1.getX(), m2.getX());
                    
                    double[] n = MatrixCalculations.multiplyVectorByX(vFromM2toM1, 1.0/distance);
                    double[] p = MatrixCalculations.sumVectors(m2.getX(), MatrixCalculations.multiplyVectorByX(vFromM2toM1, 1.0/2));
                    double[] padot = m1.getPointVelocity(p);
                    double[] pbdot = m2.getPointVelocity(p);
                    double vrel = MatrixCalculations.innerProduct(n, MatrixCalculations.subtractionOfVectors(padot, pbdot));
                    
                    
                  //   if(vrel < COLLISION_EPSILON && vrel > -1*COLLISION_EPSILON ) {
                    if(vrel == 0) {
                    //    contacts.add(new Contact(m1, m2, p, n));
                    // } else if(vrel < -1.0*COLLISION_EPSILON)
                   } else if (vrel < 0)
                        collisions.add(new Contact(m1, m2, p, n));
                }
            }
        }
        System.out.println("Tarkastus");
        if(all[0].size()>0)
          System.out.println("Trmyksi: "+all[0].size());
        if(all[1].size()>0)
        System.out.println("Kontakteja: "+all[1].size());
        return all;
    }
    
    /**
     * Typer (?) rekursiivinen tapa ajaa ensimmiseen trmykseen saakka.
     * Palauttaa ensimmisen trmyksen ajankohdan t
     */
    private double getFirstCollisionTime(double t0, double h, double [] y0) throws MatrixException {
        // System.out.println("runUntilCollision");
        if (h/2 > COLLISION_EPSILON) {
            double [] yTemp = solver.calculate(t0, h/2, y0);
            arrayToMarbles(yTemp);
            // Testataan tuliko trmyksi
            if (checkCollisions()[0].size() > 0) {
                // Trmykset ennen h:ta, jatketaan rekursiota
                return getFirstCollisionTime(t0, h/2, y0);
            } else {
                // Trmykset h:n jlkeen, jatketaan rekursiota
                return getFirstCollisionTime(t0 + h/2, h/2, yTemp);
            }
        } else {
            return t0 + h;
        }
    }
    
    public void runAnimation() throws MatrixException {
        animationRunning = true;
        double [] y0 = null;
        double [] yFinal = marblesToArray();
        double h = (double)(1.0/30.0); //1.0/30.0
        
        long time = 0;
    //    Vector all[] = new Vector[2];
        for (double t = 0.0; t < 30.0; t = t + h) { 
            System.out.println("Uusi aika-askel");
            time = System.currentTimeMillis();
            y0 = yFinal;
            yFinal = solver.calculate(t, h, y0);
            arrayToMarbles(yFinal);
            
            double tCurrent = t;
            double hCurrent = h;
            int times = 0;
            int visited =0;
            // Testataan tuliko trmyksi
            Vector[] all = checkCollisions();
           // all = checkCollisions();
            System.out.println("Menossa whileen "+ all[1].size());
           // while (all[0].size() > 0 || (all[1].size() > 0 && (times == 0 || visited == 1))) {
              while (all[0].size() > 0) {
                System.out.println("CheckCollisions -> true");
                double tc = getFirstCollisionTime(tCurrent, hCurrent, y0);
                // Ajetaan ekaan trmykseen saakka
                yFinal = solver.calculate(tCurrent, tc - tCurrent, y0);
                arrayToMarbles(yFinal);
                // Piirretn "vlivaihe"
                // Pois kun hidasti liikaa
        /*
        try {
          double timeLeft = (time + t * 1000) - System.currentTimeMillis();
          if (timeLeft > 0) {
            System.out.println(timeLeft);
            Thread.currentThread().sleep((long)timeLeft);
          }
          Thread.currentThread().sleep((long)(1000*(tc - tCurrent)));
        } catch (Exception exp) {}
        display.display();
         */
                
                // Tll hetkell trmyksess olevat
                all = checkCollisions();
                
                // Muutetaan trmvien kappaleiden voimat
                for (Enumeration e = all[0].elements(); e.hasMoreElements();) {
                    ((Contact)e.nextElement()).collision(BOUNCE_EPSILON);
                    visited = 1;
                }
                
                // Muutetaan kontaktikappaleiden voimat
                if(all[1].size()>0)
                    contact(all[1]);
                // Jatketaan aika-askel loppuun
                y0 = marblesToArray();
                hCurrent = (tCurrent + hCurrent) - tc;
                tCurrent = tc;
                yFinal = solver.calculate(tCurrent, hCurrent, y0);
                arrayToMarbles(yFinal);
                all = checkCollisions();
                times++;
            }
            display.display();
      /*
      for (int i = 0; i < getMarbleCount(); i++) {
        double[] x =  ((Marble) marbles.get(i)).getX();
        System.out.println("Kuula "+i+": "+ x[0]+" "+x[1]+" "+x[2]);
      }
       */
            //System.out.println("UUSI AIKA_ASKEL!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
            try {
                //
                double timeLeft = (time +  h*1000) - System.currentTimeMillis();
                if (timeLeft > 0) {
                    //  System.out.println(timeLeft);
                    Thread.currentThread().sleep((long)timeLeft);
                }
            } catch (Exception exp) {}
        }
        animationRunning = false;
    }
}

